#include <stdio.h>
#include <stdlib.h>
#include <libgen.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>

int FONT_SIZE = 1024;

void convert(FILE * fi, FILE * fo, int not_last) {
  int i;
  unsigned char c;

  for (i = 0; i < FONT_SIZE; i++) {
    c = fgetc(fi);

    fprintf(fo, "0x%02x", c);
    if (i < (FONT_SIZE - 1) || not_last) {
      fprintf(fo, ", ");
    }
    if (i % 16 == 15) {
      fprintf(fo, "\n");
    }
  }
}

void merge(FILE * fi1, FILE * fi2, FILE * fo, int not_last) {
  int i;
  unsigned char c1, c2, c;
  unsigned char toggle, toggle2, bit, bitval;

  toggle = toggle2 = 0;
  for (i = 0; i < FONT_SIZE; i++) {
    c1 = fgetc(fi1);
    c2 = fgetc(fi2);

    toggle2 = !toggle2;
    if (toggle2) {
      toggle = !toggle;
    }
    bitval = 128 + 64;
    c = 0;
    for (bit = 0; bit < 4; bit++) {
      if (bit % 2 == toggle) {
        c = c | (c1 & bitval);
      } else {
        c = c | (c2 & bitval);
      }
      bitval = bitval >> 2;
    }

    fprintf(fo, "0x%02x", c);
    if (i < (FONT_SIZE - 1) || not_last) {
      fprintf(fo, ", ");
    }
    if (i % 16 == 15) {
      fprintf(fo, "\n");
    }
  }
}

int main(int argc, char * argv[]) {
  FILE * fi, * fi2, * fo;
  char h_name[256], ary_name[256];
  int i, opt_merge, font_len;
  struct stat filestat;

  /* Command-line */
  if (argc < 3 || argc > 5) {
    fprintf(stderr, "Usage: %s [--merge] font1.fnt [font2.fnt] output.h\n", argv[0]);
    exit(1);
  }

  opt_merge = 0;
  if (strstr(argv[1], "--") == argv[1]) {
    if ((strcmp(argv[1], "--merge") == 0) && argc == 5) {
      opt_merge = 1;
    } else {
      fprintf(stderr, "%s --merge ... is the only valid options, & needs two fonts", argv[0]);
      exit(1);
    }
  }

  stat(argv[1 + opt_merge], &filestat);
  FONT_SIZE = filestat.st_size;
  fprintf(stderr, "Info: %s is %d bytes\n", argv[1 + opt_merge], FONT_SIZE);

  if (argc == 3) {
    font_len = FONT_SIZE;
  } else if (argc == 4) {
    font_len = FONT_SIZE * 2;
  } else if (argc == 5) {
    font_len = FONT_SIZE * 3;
  }

  /* Open output file */
  fo = fopen(argv[argc - 1], "w");
  if (fo == NULL) {
    perror(argv[argc - 1]);
    exit(1);
  }

  snprintf(h_name, sizeof(h_name), "%s", basename(argv[argc - 1]));
  snprintf(ary_name, sizeof(ary_name), "%s", basename(argv[argc - 1]));
  for (i = 0; h_name[i] != '\0'; i++) {
    if (isalnum(h_name[i])) {
      h_name[i] = toupper(h_name[i]);
    } else {
      if (h_name[i] != '.') {
        ary_name[i] = '_';
      } else {
        ary_name[i] = '\0';
      }
      h_name[i] = '_';
    }
  }

  fprintf(fo, "#ifndef %s\n", h_name);
  fprintf(fo, "#define %s\n", h_name);
  fprintf(fo, "/* generated by tools/fonts-to-h */\n");
  fprintf(fo, "#define %s_LEN %d\n", h_name, font_len);
  fprintf(fo, "static unsigned char %s[] = {\n", ary_name);

  fi = fopen(argv[1 + opt_merge], "rb");
  if (fi == NULL) {
    perror(argv[1 + opt_merge]);
    exit(1);
  }
  fprintf(fo, "/* %s */\n", argv[1 + opt_merge]);
  convert(fi, fo, 1);
  fclose(fi);

  if (argc > 3) {
    fi = fopen(argv[2 + opt_merge], "rb");
    if (fi == NULL) {
      perror(argv[2 + opt_merge]);
      exit(1);
    }
    fprintf(fo, "/* %s */\n", argv[2 + opt_merge]);
    convert(fi, fo, 1);
    fclose(fi);

    if (opt_merge) {
      fi = fopen(argv[1 + opt_merge], "rb");
      if (fi == NULL) {
        perror(argv[1 + opt_merge]);
        exit(1);
      }
      fi2 = fopen(argv[2 + opt_merge], "rb");
      if (fi2 == NULL) {
        perror(argv[2 + opt_merge]);
        exit(1);
      }
      fprintf(fo, "/* %s + %s combined */\n", argv[1 + opt_merge], argv[2 + opt_merge]);
      merge(fi, fi2, fo, 0);
      fclose(fi);
      fclose(fi2);
    }
  }

  fprintf(fo, "};\n");
  fprintf(fo, "#endif /* %s */\n", h_name);

  fclose(fo);

  return(0);
}
